#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov  9 16:27:41 2016
# Last Modified : lun. 09 avril 2018 21:56:41 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
from matplotlib.pyplot import figure, plot, imshow, show
import numpy as np
import os
from time import time, sleep
from multiprocessing import Process
import subprocess
import sys
import locale
from inspect import getsource
from cpyvke.curseswin.widgets import suspend_curses
from cpyvke.utils.comm import send_msg


locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


def get_source_code(code):
    return getsource(code)


def inspect_class_instance(class_inst):
    """ Return a dictionnary of :
        * class attributes    : inst.__class__.__dict__
        * instance attributes : int.__dict__
    """
    class_attr = {i: {'type': str(type(class_inst.__class__.__dict__[i])),
                      'value': str(class_inst.__class__.__dict__[i])} for i in
                  class_inst.__class__.__dict__ if not i.startswith('_')}

    inst_attr = {i: {'type': '@' + str(type(class_inst.__dict__[i])),
                     'value': str(class_inst.__dict__[i])} for i in
                 class_inst.__dict__ if not i.startswith('_')}

    return dict(class_attr, **inst_attr)


def inspect_class(class_inst):
    """ Return a dictionnary of :
        * class attributes    : inst.__class__.__dict__
    """
    class_attr = {i: {'type': str(type(class_inst.__dict__[i])),
                      'value': str(class_inst.__dict__[i])} for i in
                  class_inst.__dict__ if not i.startswith('_')}

    return class_attr


def threaded(func):
    """ Run func in thread """

    def run(*k, **kw):
        t = Process(target=func, args=k, kwargs=kw)
        t.start()
        return t

    return run


class Inspect:
    """ Plot/Display variable. """

    def __init__(self, varval, varname, vartype):

        self.varval = varval
        self.vartype = vartype
        self.varname = varname

    @staticmethod
    def type_struct():
        return ('dict', 'tuple', 'list', 'set', 'frozenset',
                'str', 'unicode', 'bytes', 'bytearray')

    @staticmethod
    def type_numeric():
        return ('int', 'float', 'complex')

    @threaded
    def plot2D(self):
        """ Plot 2D variable. """

        figure()
        imshow(self.varval)
        show()

    @threaded
    def plot1D(self):
        """ Plot 1D variable. """

        figure()
        plot(self.varval)
        show()

    @threaded
    def plot1Dcols(self):
        """ Plot 2D variable : superimpose all columns """

        figure()
        for i in range(np.shape(self.varval)[1]):
            plot(self.varval[:, i])
        show()

    @threaded
    def plot1Dlines(self):
        """ Plot 2D variable : superimpose all lines """

        figure()
        for i in range(np.shape(self.varval)[0]):
            plot(self.varval[i, :])
        show()

    def save_np(self, varname, SaveDir, METHOD='npy'):
        """ Save numpy variable to file """

        if METHOD == 'txt':
            np.savetxt(SaveDir + varname + '.txt', self.varval)

        if METHOD == 'npy':
            np.save(SaveDir + varname, self.varval)

        if METHOD == 'npz':
            np.savez_compressed(SaveDir + varname, var=self.varval)

    def save(self, SaveDir):
        """ Save variable to text file """

        filename = SaveDir + self.varname

        if self.vartype != 'str' and self.vartype != 'unicode':
            self.varval = str(self.varval)

        if self.vartype == 'unicode':
            with open(filename, 'w') as f:
                f.write(self.varval.encode(code))

        else:
            with open(filename, 'w') as f:
                f.write(self.varval)

    def display(self, app='less', arg='Help'):
        """
        Display variable using **App** (less|vim).
        If a module is provided, use **arg** to choose wich methode to use.
        """

        filename = '/tmp/tmp_cVKE'

        # Convert all type of variable to string
        if self.vartype != 'str' and self.vartype != 'unicode':
            self.varval = str(self.varval)

        #
        if self.vartype == 'unicode':
            with open(filename, 'w') as f:
                f.write(self.varval.encode(code))

        elif self.vartype == 'function':
            if arg == 'help':
                with open(filename, 'w') as f:
                    f.write(self.varval)

        elif self.vartype == 'module':
            if arg == 'help':
                sys.stdout = open(filename, 'w')
                try:
                    exec('import {}'.format(self.varval))
                    print(help(self.varval))
                except ImportError:
                    print('Not found!')

        else:
            with open(filename, 'w') as f:
                f.write(self.varval)

        with suspend_curses():
            subprocess.run([app, filename])
            subprocess.run(['rm', filename])


class ProceedInspection:
    """ Object inspection """

    def __init__(self, app, sock, logger, name, value, typ, pos, page):
        """ Class constructor """

        self.app = app
        self.sock = sock
        self.logger = logger
        self.varname = name
        self.varval = value
        self.vartype = typ
        self.position = pos
        self.page = page
        self.doc = None

    def get_variable(self):
        """ Get Variable characteristics. """
        if self.vartype in ['module']:
            self.get_module()

        if self.vartype in ['function']:
            self.get_function_code()
            self.get_function_doc()

        elif self.vartype in ['builtin_function_or_method']:
            self.get_function_doc()

        elif self.vartype in Inspect.type_struct():
            self.get_structure()

        elif self.vartype == 'ndarray':
            self.get_ndarray()

        elif '.' + self.vartype in self.varval:     # Class instance
            self.vartype = 'class'
            self.get_class_instance()

        elif self.vartype == 'type':                # Class
            self.get_class()

        else:
            self._ismenu = True

        return self._ismenu, self.varname, self.varval, self.vartype, self.doc

    def get_function_doc(self):
        """ Get function __doc__ """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str({}.__doc__))".format(self.filename, self.varname)
        try:
            send_msg(self.sock.RequestSock, '<code>' + self.code)
            self.logger.debug("Inspecting '{}' with type '{}'".format(self.varname, self.vartype))
            self.wait()
            with open(self.filename, 'r') as f:
                self.doc = f.read()
        except Exception:
            self.logger.error('Get traceback', exc_info=True)
            self.kernel_busy()
        else:
            self.logger.debug('kd5 answered : {}'.format(self.varval))
            os.remove(self.filename)
            self._ismenu = True

    def get_function_code(self):
        """ Get function __doc__ """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str(_inspect.getsource({})))".format(self.filename, self.varname)
        self.send_code()

    def get_class_instance(self):
        """ Get Class characteristics. """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str(_inspect.inspect_class_instance({})))".format(self.filename, self.varname)
        self.send_code()

    def get_class(self):
        """ Get Class characteristics. """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str(_inspect.inspect_class({})))".format(self.filename, self.varname)
        self.send_code()

    def get_module(self):
        """ Get modules characteristics """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str({}.__name__))".format(self.filename, self.varname)
        self.send_code()

    def get_structure(self):
        """ Get Dict/List/Tuple characteristics """

        self.filename = '/tmp/tmp_' + self.varname
        self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str({}))".format(self.filename, self.varname)
        self.send_code()

    def get_ndarray(self):
        """ Get ndarray characteristics. """

        self.filename = '/tmp/tmp_' + self.varname + '.npy'
        code = "_np.save('" + self.filename + "', " + self.varname + ')'
        try:
            send_msg(self.sock.RequestSock, '<code>' + code)
            self.logger.debug("Name of module '{}' asked to kd5".format(self.varname))
            self.wait()
            self.varval = np.load(self.filename)
        except Exception:
            self.logger.error('Get traceback : ', exc_info=True)
            self.kernel_busy()
        else:
            self.logger.debug('kd5 answered')
            os.remove(self.filename)
            self._ismenu = True

    def get_help(self):
        """ Help item in menu """

        if self.vartype == 'function':
            self.filename = '/tmp/tmp_' + self.varname
            self.code = "with open('{}' , 'w') as fcpyvke0:\n\tfcpyvke0.write(str({}.__doc__))".format(self.filename, self.varname)
            self.send_code()

    def send_code(self):
        """ Send code to kernel and except answer ! """

        try:
            send_msg(self.sock.RequestSock, '<code>' + self.code)
            self.logger.debug("Inspecting '{}' with type '{}'".format(self.varname, self.vartype))
            self.wait()
            with open(self.filename, 'r') as f:
                if self.vartype in ['str', 'function', 'module', 'builtin_function_or_method']:
                    self.varval = f.read()
                else:
                    self.varval = eval(f.read())
        except Exception:
            self.logger.error('Get traceback', exc_info=True)
            self.kernel_busy()
        else:
            self.logger.debug('kd5 answered : {}'.format(self.varval))
            os.remove(self.filename)
            self._ismenu = True

    def kernel_busy(self):
        """ Handle silent kernel. """

        self.app.wng.display('Kernel Busy ! Try again...')
        self.varval = '[Busy]'
        self._ismenu = False

    def wait(self):
        """ Wait for variable value. """

        i, j = 0, 0
        spinner = [['.', 'o', 'O', 'o'],
                   ['.', 'o', 'O', '@', '*'],
                   ['v', '<', '^', '>'],
                   ['(o)(o)', '(-)(-)', '(_)(_)'],
                   ['◴', '◷', '◶', '◵'],
                   ['←', '↖', '↑', '↗', '→', '↘', '↓', '↙'],
                   ['▁', '▃', '▄', '▅', '▆', '▇', '█', '▇', '▆', '▅', '▄', '▃'],
                   ['▉', '▊', '▋', '▌', '▍', '▎', '▏', '▎', '▍', '▌', '▋', '▊', '▉'],
                   ['▖', '▘', '▝', '▗'],
                   ['▌', '▀', '▐', '▄'],
                   ['┤', '┘', '┴', '└', '├', '┌', '┬', '┐'],
                   ['◢', '◣', '◤', '◥'],
                   ['◰', '◳', '◲', '◱'],
                   ['◐', '◓', '◑', '◒'],
                   ['|', '/', '-', '\\'],
                   ['.', 'o', 'O', '@', '*'],
                   ['◡◡', '⊙⊙', '◠◠'],
                   ['◜ ', ' ◝', ' ◞', '◟ '],
                   ['◇', '◈', '◆'],
                   ['⣾', '⣽', '⣻', '⢿', '⡿', '⣟', '⣯', '⣷'],
                   ['⠁', '⠂', '⠄', '⡀', '⢀', '⠠', '⠐', '⠈'],
                   ['Searching.', 'Searching..', 'Searching...']
                   ]

        search = spinner[21]
        spinner = spinner[19]
        ti = time()
        while os.path.exists(self.filename) is False:
            self.app.stdscr.addstr(self.position + 1 - (self.page-1)*self.app.row_max, 1,
                                   spinner[i], self.app.c_exp_txt | curses.A_BOLD)
            self.app.stdscr.addstr(self.app.screen_height - 1, 0,
                                   search[j], self.app.c_exp_txt | curses.A_DIM)

            self.app.stdscr.refresh()
            if i < len(spinner) - 1:
                i += 1
            else:
                i = 0

            if j < len(search) - 1:
                j += 1
            else:
                j = 0

            if time() - ti > 3:
                break

            sleep(0.05)

        self.app.stdscr.refresh()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mar. 13 mars 2018 12:01:45 CET
# Last Modified : ven. 30 mars 2018 23:48:03 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import textwrap


def str_reduce(msg, maxwidth):
    """ Reduce a string with dots (...) if its length is greater than maxwidth """

    if len(msg) >= maxwidth:
        return msg[:maxwidth-3] + '...'
    else:
        return msg


def str_format(string, width):
    """ Convert a string into a list whose elements length do not exceed a given width. """

    output = []
    for seq in string.split('\n'):
        tmp = textwrap.wrap(seq, width)
        for split in tmp:
            output.append(split)

    return output


def whos_to_dic(string):
    """ Format output of daemon to a dictionnary """

    variables = {}
    for item in string.split('\n')[2:]:
        tmp = [j for j in item.split(' ') if j is not '']
        if tmp:
            var_name = tmp[0]
            var_typ = tmp[1]
            var_val = ' '.join(tmp[2:])
            variables[var_name] = {'value': var_val, 'type': var_typ}
    return variables


def dump(obj, nested_level=0, output=[]):
    """ Format dict, list and tuples variables for displaying. """

    if nested_level == 0:
        output = []

    spacing = '   '
    if type(obj) is dict:
        output.append('%s{' % ((nested_level) * spacing))
        for k, v in obj.items():
            if hasattr(v, '__iter__'):
                output.append('%s%s:' % ((nested_level + 1) * spacing, k))
                dump(v, nested_level + 1, output)
            else:
                output.append('%s%s: %s' % ((nested_level + 1) * spacing, k, v))
        output.append('%s}' % (nested_level * spacing))

    elif type(obj) is list:
        output.append('%s[' % ((nested_level) * spacing))
        for v in obj:
            if hasattr(v, '__iter__'):
                dump(v, nested_level + 1, output)
            else:
                output.append('%s%s' % ((nested_level + 1) * spacing, v))
        output.append('%s]' % ((nested_level) * spacing))

    elif type(obj) is tuple:
        output.append('%s(' % ((nested_level) * spacing))
        for v in obj:
            if hasattr(v, '__iter__'):
                dump(v, nested_level + 1, output)
            else:
                output.append('%s%s' % ((nested_level + 1) * spacing, v))
        output.append('%s)' % ((nested_level) * spacing))

    elif type(obj) in [set, frozenset]:
        output.append('%s{' % ((nested_level) * spacing))
        for v in obj:
            if hasattr(v, '__iter__'):
                dump(v, nested_level + 1, output)
            else:
                output.append('%s%s' % ((nested_level + 1) * spacing, v))
        output.append('%s}' % ((nested_level) * spacing))

    else:
        output.append('%s%s' % (nested_level * spacing, obj))

    return output


def format_cell(variables, name, screen_width):
    """ Format data for display """

    screen_width = screen_width - 4

    if '/kernel' in variables[name]['value']:
        return format_kernel(variables, name, screen_width)
    else:
        return format_variable(variables, name, screen_width)


def format_variable(variables, name, screen_width):
    """ Format regular variables """

    max_width = int(screen_width/5)

    # Types
    if "@<class '" in variables[name]['type']:
        typ = '[inst. ' + variables[name]['type'].split("'")[1] + ']'
    elif "<class '" in variables[name]['type']:
        typ = '[' + variables[name]['type'].split("'")[1] + ']'
    else:
        typ = '[' + variables[name]['type'] + ']'

    # Only display module name
    if 'module' in variables[name]['value']:
        val = variables[name]['value'].split("'")[1]

    # Only display dimensions of array
    elif 'elem' in variables[name]['value']:
        val1 = variables[name]['value'].split(":")[0]
        val2 = variables[name]['value'].split("`")[1]
        val = val1 + ' [' + val2 + ']'

    elif 'classmethod' in variables[name]['value']:
        val = 'classmethod'

    elif 'staticmethod' in variables[name]['value']:
        val = 'staticmethod'

    elif ' at' in variables[name]['value']:
        val = variables[name]['value'].split(' at ')[0] + '>'

    elif 'frozenset' in variables[name]['value']:
        val = variables[name]['value'].split('(')[1].split(')')[0]

    elif '<...>' in variables[name]['value']:
        val = variables[name]['value'].split('<')[0] + '...'

    else:
        # Repr to avoid interpreting \n in strings
        val = variables[name]['value']

    # Check length of each entry
    if len(val) > 3*max_width:
        val = val[0:3*max_width-4] + '... '

    if len(name) > max_width:
        name = name[0:max_width-4] + '... '

    if len(typ) > max_width:
        typ = typ[0:max_width-4] + '... '

    s1 = "{:{wname}} {:{wval}}".format(name, val, wname=max_width, wval=3*max_width)
    s2 = "{:{wtype}}".format(typ, wtype=screen_width-len(s1))

    return s1, s2


def format_kernel(variables, name, screen_width):
    """ Format regular variables """

    max_width = int(screen_width/5)
    typ = '[' + variables[name]['type'] + ']'

    # Repr to avoid interpreting \n in strings
    val = variables[name]['value']

    # Check length of each entry
    if len(val) > 3*max_width:
        val = val[0:3*max_width-4] + '... '

    if len(name) > max_width:
        name = name[0:max_width-4] + '... '

    if len(typ) > max_width:
        typ = typ[0:max_width-4] + '... '

    s1 = "{:{wname}} {:{wval}}".format(name, val, wname=max_width, wval=3*max_width)
    s2 = "{:{wtype}}".format(typ, wtype=screen_width-len(s1))

    return s1, s2



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date :
# Last Modified : mer. 28 mars 2018 17:20:28 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


import os


def main():
    """
    launch ipython console.
    """
    logdir = os.path.expanduser('~') + '/.cpyvke/'
    lockfile = logdir + 'kd5.lock'

    if os.path.exists(lockfile):
        with open(lockfile, 'r') as f:
            kid = f.readline()

        cmd = 'jupyter console --existing kernel-{}.json'.format(kid)
        os.system(cmd)

    else:
        print('No associated kernel found !')


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Fri Nov  4 21:49:15 2016
# Last Modified : lun. 09 avril 2018 21:32:29 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import sys
import threading
import socket
import argparse
import logging
import psutil
from time import sleep
from queue import Queue
from logging.handlers import RotatingFileHandler
from jupyter_client import find_connection_file

from .utils.kernel import init_kernel, connect_kernel, print_kernel_list, \
    start_new_kernel, set_kid
from .utils.kd import is_kd_running, find_lost_pid, kdwrite, kdread
from .utils.comm import send_msg, recv_msg
from .utils.daemon3x import Daemon
from .utils.config import cfg_setup
from .utils.term_colors import RED, BLUE, CYAN, RESET

logger = logging.getLogger('kd5')


class Watcher(threading.Thread):
    """
    Daemon : watch the kernel input and update variable list.
    The client may also request for the content of a variable.
    """

    def __init__(self, kc, delay=0.1, sport=15555, rport=15556):
        """ Class constructor """

        # Init Thread
        threading.Thread.__init__(self)
        threading.Thread.daemon = True

        # Events
        self._stop = threading.Event()
        self._pause = threading.Event()

        # Inputs
        self.kc = kc
        self.delay = delay

        # Queue
        self.kernel_queue = Queue()

        # Init Main Socket
        self.MainSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.MainSock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.MainSock.bind(('', sport))
        self.MainSock.listen(5)
        self.MainSock.setblocking(0)
        self.client_main = None

        # Init Request Socket
        self.RequestSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.RequestSock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.RequestSock.bind(('', rport))
        self.RequestSock.listen(5)
        self.RequestSock.setblocking(0)
        self.client_request = None

        logger.info('++++++++++++++++++++++++++++')
        logger.info('Daemon started !')
        logger.info('Delay set to {} s.'.format(self.delay))
        logger.info('Kernel : {}'.format(self.kc.connection_file))
        logger.info('Streaming on {}'.format(sport))
        logger.info('Listening on {}'.format(rport))
        logger.info('++++++++++++++++++++++++++++')

        # Init variables
        self.msg = 0
        self.check_input()
        self.variables = ''

    def run(self):
        """ Run the variable explorer daemon """

        # Launch streamer
        streamer = threading.Thread(target=self.stream_data)
        streamer.start()

        while True:

            # Look for connection to request socket
            self.listen_request_sock()

            # Watch RequestSock for code request OR kernel changes
            if self.client_request:
                self.fetch_request()

            # Terminate daemon
            if self._stop.isSet():
                logger.info('Request Socket closed !')
                break

            sleep(self.delay)

        streamer.join()
        # Close connection to clients
        self.MainSock.shutdown()
        self.RequestSock.shutdown()
        # Destroy socket
        self.MainSock.close()
        self.RequestSock.close()
        logger.info('Exited')

    def stream_data(self):
        """ Stream 'whos' output to client. """

        while True:

            # Pause if client has a request
            self.pause()

#            # Check if reset has been invoked
#            if 'empty' in self.variables:
#                init_kernel(self.kc)

            # Look for connection to main socket
            self.listen_main_sock()

            # Listen to kernel changes
            if self.kernel_queue.qsize() > 0:
                self.kernel_change(self.kernel_queue.get())

            # Check in new entries in kernel
            self.check_input()

            # If new entries, update variables
            if self.msg == 1:
                self.send_variables()

            # Terminate daemon
            if self._stop.isSet():
                logger.info('Stream Socket closed !')
                break

            sleep(self.delay)

    def pause(self):
        """ Pause streamer when client has a request. """

        if self._pause.isSet():

            logger.debug('Streamer paused...')

            while self._pause.isSet():
                sleep(self.delay)

            logger.debug('Streamer active')

            # Force Update var list
            self.send_variables()

    def check_input(self):
        """ Check the iopub msgs available """

        self.msg = 0
        while self.kc.iopub_channel.msg_ready():
            data = self.kc.get_iopub_msg(timeout=0.1)
            logger.debug('WATCHING : {}'.format(self.disp_data(data)))
            self.msg = 1
            self.check_init(data)

    def check_init(self, data):
        if 'code' in data['content'].keys():
            if 'reset' in data['content']['code']:
                init_kernel(self.kc)
                logger.debug('RESET RECEIVED : {}'.format('Init Kernel'))

    def execute(self, code):
        """ Execute **code** """

        value = None
        MSG_RECEIVED = False

        msg_id = self.kc.execute(code, store_history=False)
        logger.debug("EXEC : '{}' sent with id {}".format(code, msg_id.split('-')[0]))

        while not MSG_RECEIVED:
            self.wait_msg()

            while self.kc.iopub_channel.msg_ready():
                data = self.kc.get_iopub_msg()
                if data['parent_header']['msg_id'] != msg_id:
                    logger.debug('EXEC : PASS MSG : {}'.format(self.disp_data(data)))
                    self.check_init(data)
                    continue
                else:
                    MSG_RECEIVED = True
                    logger.debug('EXEC : PROCEED MSG : {}'.format(self.disp_data(data)))
                    if data['header']['msg_type'] == 'stream':
                        value = data['content']['text']

        logger.debug('EXEC : RESULT :\n {}'.format(value))
        self.msg = 0

        return value

    def wait_msg(self):
        """ Waiting for message from iopub channel """

        while not self.kc.iopub_channel.msg_ready():  # To fix.Have to wait
            sleep(self.delay)

    def listen_main_sock(self):
        """ Look for client connection to main socket. """

        try:
            self.client_main, address = self.MainSock.accept()
            logger.info("{} connected to main socket".format(address))
        except BlockingIOError:
            pass
        else:
            send_msg(self.client_main, self.variables)

    def listen_request_sock(self):
        """ Look for client connection to request socket. """

        try:
            self.client_request, address = self.RequestSock.accept()
            logger.info("{} connected to request socket".format(address))
        except BlockingIOError:
            pass

    def check_variables(self):
        """ If variables is None, ask again to kernel """
        if not self.variables:
            logger.debug("EXEC : 'variables' is None : RUN AGAIN 'whos'")
        while not self.variables:
            self.variables = self.execute('whos')

    def kernel_change(self, cf):
        """ Watch kernel changes """

        old_id = set_kid(self.kc.connection_file)
        _, self.kc = connect_kernel(cf)
        new_id = set_kid(self.kc.connection_file)

        # Update kd5.lock files
        self.update_lockfile(new_id)
        logger.info('Kernel change from {} to {}'.format(old_id, new_id))

        # Force kernel update
        self.send_variables()

    def update_lockfile(self, new_id):
        """ Update lock files """

        LogDir = os.path.expanduser("~") + "/.cpyvke/"

        with open(LogDir + 'kd5.lock', 'w') as f:
            f.write(new_id)

    def send_variables(self):
        """ Send variable to client """

        self.variables = self.execute('whos')
        self.check_variables()
        # Send to client
        if self.client_main:
            try:
                send_msg(self.client_main, self.variables)
            except BlockingIOError:
                logger.info("Client is disconnected from main socket!")
                self.client_main = None
            else:
                logger.info('Variable list sent to client')

    @staticmethod
    def disp_id(data):
        """ Display first seq of message id only """
        return data['parent_header']['msg_id'].split('-')[0]

    @classmethod
    def disp_data(cls, data):
        if data['msg_type'] == 'status':
            dbg = '{} | status : {}'.format(cls.disp_id(data), data['content']['execution_state'])
        elif data['msg_type'] == 'execute_input':
            dbg = '{} | code : {}'.format(cls.disp_id(data), data['content']['code'])
        elif data['msg_type'] == 'stream':
            dbg = '{} | stream'.format(cls.disp_id(data))
        elif data['msg_type'] == 'error':
            dbg = '{} | error : {}'.format(cls.disp_id(data), data['content']['ename'])
        else:
            dbg = '{} {}'.format(cls.disp_id(data), data['msg_type'])
        return dbg

    def fetch_request(self):
        """ Listen to sock request :
            handle kernel changes | exec code | stop signal. """

        try:
            tmp = recv_msg(self.client_request).decode('utf8')
        except AttributeError:
            tmp = None
            self.client_request = None
            logger.info("Client is disconnected from request socket!")

        if tmp:
            self._pause.set()
            sleep(self.delay)
            logger.info('Request from client')
            logger.debug('RECEIVED :\n {}'.format(tmp))

            if '<cf>' in tmp:
                cf = tmp.split('<cf>')[1]
                self.kernel_queue.put(cf)           # Send new k to streamer

            elif '<_stop>' in tmp:
                self.stop()

            elif '<code>' in tmp:
                self.execute(tmp.split('<code>')[1])

            self._pause.clear()

    def stop(self):
        """ Stop thread. """

        logger.info("Client sent SIGTERM")
        self._stop.set()


class Daemonize(Daemon):
    """ Daemonize a class """

    def __init__(self,
                 pidfile,
                 WatcherArgs,
                 stdin='/dev/null',
                 stdout='/dev/null',
                 stderr='/dev/null'):
        """ Override __init__ Daemon method with this method. """

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile
        self.cf = WatcherArgs['cf']
        self.sport = WatcherArgs['sport']
        self.rport = WatcherArgs['rport']
        self.delay = WatcherArgs['delay']

    def run(self):
        """ Override Daemon run method with this method. """

        km, kc = connect_kernel(self.cf)
        WK = Watcher(kc,
                     sport=self.sport,
                     rport=self.rport,
                     delay=self.delay)
        WK.start()
        WK.join()


def parse_args(lockfile, pidfile, Config):
    """ Parse arguments. """

    parser = argparse.ArgumentParser()
    parser.add_argument('action', choices=('start', 'stop', 'restart',
                                           'last', 'status', 'list'))
    parser.add_argument("integer",
                        help="Start up with existing kernel. \
                        INTEGER is the id of the connection file.",
                        nargs='?')
    args = parser.parse_args()

    # start|last actions
    if args.action in ['start', 'last']:
        current_pid = status_pid(pidfile)
        if current_pid:
            sys.stderr.write("Cannot start new instance of kd5.\n")
            sys.exit(1)
        else:
            kid = start_action(args, lockfile, Config)

    # restart action
    elif args.action == 'restart':
        kid = restart_action(pidfile, lockfile)

    # Stop action
    elif args.action == 'stop':
        kid = stop_action(pidfile, lockfile)

    # List action
    elif args.action == 'list':
        print_kernel_list()
        sys.exit(0)

    # Status action
    elif args.action == 'status':
        status_action(pidfile, lockfile)
        sys.exit(0)

    return args, kid


def create_new(Config):
    """ Create new kernel """

    sys.stdout.write('Creating new kernel...\n')
    kid = start_new_kernel(version=Config['kernel version']['version'])
    message = 'Kernel id {} created (Python {})\n'
    sys.stdout.write(message.format(kid, Config['kernel version']['version']))

    return kid


def start_action(args, lockfile, Config):
    """ Start Parser action. """

    if args.integer:

        try:
            kid = str(args.integer)
            find_connection_file(kid)
        except OSError:
            message = '{}Error :\t{}Cannot find kernel id. {} !\n\tExiting\n'
            sys.stderr.write(message.format(RED, RESET, args.integer))
            sys.exit(1)
        else:
            message = 'Connecting to kernel id. {}\n'
            sys.stdout.write(message.format(kid))

    elif args.action == 'last':
        kid = kdread(lockfile)
        if kid:
            message = 'Connecting to kernel id. {}\n'
            sys.stdout.write(message.format(kid))
        else:
            sys.exit(1)

    else:
        kid = create_new(Config)

    kdwrite(lockfile, kid)

    return kid


def restart_action(pidfile, lockfile):
    """ Prepare restart action """

    current_pid = status_pid(pidfile)
    kid = status_lock(lockfile)
    if current_pid and kid:
        return kid
    else:
        sys.stdout.write('Cannot restart kd5.\n')
        sys.exit(1)


def stop_action(pidfile, lockfile):
    """ Prepare stop action """

    current_pid = status_pid(pidfile)
    kid = status_lock(lockfile)
    if current_pid:
        if kid:
            message = 'Disconnecting from kernel id. {}\n'
            sys.stdout.write(message.format(kid))
            return kid
        else:
            message = 'SIGTERM sent to kd5\n'
            p = psutil.Process(int(current_pid))
            p.terminate()
            os.remove(pidfile)
            sys.stdout.write(message.format(kid))
            sys.exit(1)
    else:
        sys.exit(1)


def status_action(pidfile, lockfile):

    status_pid(pidfile)
    status_lock(lockfile)


def status_pid(pidfile):
    """ Status of the daemon """

    if os.path.exists(pidfile):
        if is_kd_running(pidfile):
            message = "{}Current state : {}running{} with pid {}!\n"
            sys.stderr.write(message.format(CYAN, BLUE, RESET, kdread(pidfile)))
            return kdread(pidfile)
        else:
            message = "{}Current state : {}stopped{}, but pidfile still exists !\nFix : {} deleted !\n"
            sys.stderr.write(message.format(CYAN, RED, RESET, pidfile))
            os.remove(pidfile)
            return None
    else:
        pids = find_lost_pid()
        if len(pids) == 1:
            message = "{}Current state : {}running{} with pid {}, but pidfile has been removed !\nFix : {} created\n"
            sys.stderr.write(message.format(CYAN, BLUE, RESET, pids[0], pidfile))
            with open(pidfile, 'w') as f:
                f.write(str(pids[0]))
            return pids[0]
        elif len(pids) == 0:
            message = "{}Current state : {}stopped{} \n"
            sys.stderr.write(message.format(CYAN, RED, RESET))
            return None
        elif len(pids) > 1:
            message = "{}Current state : {}multiple instances of kd5 are running !\nExiting ..."
            sys.stderr.write(message.format(CYAN, RESET))
            sys.exit(1)


def status_lock(lockfile):
    """ Information about the kernel to which the daemon is connected """

    if os.path.exists(lockfile):
        return kdread(lockfile)
    else:
        sys.stderr.write('Kernel id. not found !\n')
        return None


def main(args=None):
    """ Main """

    cfg = cfg_setup()
    Config = cfg.run()

    logdir = os.path.expanduser('~') + '/.cpyvke/'
    logfile = logdir + 'kd5.log'
    lockfile = logdir + 'kd5.lock'
    pidfile = logdir + 'kd5.pid'

    # Logger
    logger.setLevel(logging.DEBUG)

    # create the logging file handler
    handler = RotatingFileHandler(logfile, maxBytes=10*1024*1024,
                                  backupCount=5)
    logmsg = '%(asctime)s :: %(name)s :: %(threadName)s :: %(levelname)s :: %(message)s'
    formatter = logging.Formatter(logmsg, datefmt='%Y-%m-%d - %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Parse Arguments
    args, kid = parse_args(lockfile, pidfile, Config)

    sport = int(Config['comm']['s-port'])
    rport = int(Config['comm']['r-port'])
    delay = float(Config['daemon']['refresh'])

    try:
        cfile = find_connection_file(kid)
    except OSError:
        sys.stderr.write('Lockfile points to an unknown connection file !\n')
        kid = create_new(Config)
        cfile = find_connection_file(kid)
        kdwrite(lockfile, kid)

    WatchConf = {'cf': cfile,
                 'delay': delay,
                 'sport': sport,
                 'rport': rport}

    daemon = Daemonize(pidfile, WatchConf, stdout=logfile, stderr=logfile)

    if args.action == 'stop':
        daemon.stop()
        sys.stdout.write('kd5 stopped !\n')
    elif args.action == 'start' or args.action == 'last':
        sys.stdout.write('kd5 starting...\n')
        daemon.start()
    elif args.action == 'restart':
        sys.stdout.write('kd5 restarting...\n')
        daemon.restart()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Mon Nov 14 09:08:25 2016
# Last Modified : lun. 02 avril 2018 13:27:47 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
import locale
import abc


code = locale.getpreferredencoding()


class PadWin(abc.ABC):
    """ General Pad Abstract Class """

    def __init__(self, app):

        self.app = app

    @abc.abstractmethod
    def color(self, item):
        """ Pad colors """

    @property
    @abc.abstractmethod
    def title(self):
        """ Pad Title. Must be a string """

    @property
    @abc.abstractmethod
    def content(self):
        """ Pad content. Must return a list ! """

    def init_pad(self):
        """ Init Pad """

        self.pad_width = max(len(self.title),
                             max([len(elem) for elem in self.content])) + 8
        self.pad_width = min(self.pad_width, self.app.screen_width-2)
        self.pad_height = len(self.content) + 2
        self.gpad = curses.newpad(self.pad_height, self.pad_width)
        self.gpad.keypad(1)
        self.gpad.bkgd(self.color('txt'))
        self.gpad.attrset(self.color('bdr'))
        self.gpad.border(0)
        self.title_loc = int((self.pad_width - len(self.title) - 2)/2)

        # Pad content
        for i, j in enumerate(self.content):
            self.gpad.addstr(1+i, 3, j.encode(code), self.color('txt'))

        # Pad title
        if self.app.config['font']['pw-font'] == 'True':
            self.gpad.addstr(0, self.title_loc, '', self.color('pwf'))
            self.gpad.addstr(self.title, self.color('ttl'))
            self.gpad.addstr('', self.color('pwf'))
        else:
            self.gpad.addstr(0, self.title_loc, '|' + self.title + '|',
                             self.color('ttl'))

    def display(self):
        """ Create pad to display variable content. """

        self.init_pad()

        pkey = -1
        pad_pos = 0
        pad_y = max(self.pad_height,
                    self.app.screen_height - 2) - (self.app.screen_height - 2)

        while pkey not in self.app.kquit:
            if pkey in self.app.kdown:
                pad_pos = min(pad_y, pad_pos+1)
            elif pkey in self.app.kup:
                pad_pos = max(0, pad_pos-1)
            elif pkey in self.app.kright:
                pad_pos = min(pad_y, pad_pos+5)
            elif pkey in self.app.kleft:
                pad_pos = max(0, pad_pos-5)
            elif pkey == curses.KEY_NPAGE:
                pad_pos = min(pad_y, pad_pos+10)
            elif pkey == curses.KEY_PPAGE:
                pad_pos = max(0, pad_pos-10)
            elif pkey == 262:               # Home
                pad_pos = 0
            elif pkey == 360:               # End
                pad_pos = pad_y
            elif pkey == curses.KEY_RESIZE:
                break

            self.gpad.refresh(pad_pos, 0, 1, 1,
                              self.app.screen_height-2, self.app.screen_width-2)

            pkey = self.gpad.getch()

        self.gpad.erase()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov  9 10:03:04 2016
# Last Modified : mar. 10 avril 2018 20:54:45 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import sys
import locale
import logging
import argparse
import time
from jupyter_client import find_connection_file
from logging.handlers import RotatingFileHandler

from .curseswin.app import InitApp
from .curseswin.mainwin import MainWin
from .utils.config import cfg_setup
from .utils.kernel import connect_kernel, print_kernel_list
from .utils.kd import kd_status
from .utils.sockets import SocketManager
from .utils.term_colors import RED, RESET

locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


def init_cf(lockfile):
    """ Init connection file. """

    with open(lockfile, 'r') as f:
        kid = f.readline()

    return find_connection_file(kid)


def with_daemon(lockfile, pidfile, cmd):
    """ Launch daemon. """

    os.system(cmd)

    while os.path.exists(pidfile) is False:
        time.sleep(0.1)

    return init_cf(lockfile)


def no_lock_exit():
    """ If no kd5.lock ! """

    message = '{}Error :{}\tCannot find kd5.lock !\n\tFixing issues shutting down kd5...\n'
    sys.stderr.write(message.format(RED, RESET))
    os.system('kd5 stop')
    sys.stderr.write("You can now restart cpyvke!\n")
    sys.exit(1)


def parse_args(lockfile, pidfile):
    """ Parse Arguments. """

    parser = argparse.ArgumentParser()
    parser.add_argument("-L", "--list", help="List all kernels",
                        action="store_true")
    parser.add_argument("integer", help="Start up with existing kernel. \
                        INTEGER is the id of the connection file. \
                        INTEGER can also be the keyword 'last' for 'last kernel'",
                        nargs='?')

    args = parser.parse_args()

    pid = kd_status(pidfile)

    if args.list:
        print_kernel_list()
        sys.exit(0)

    elif os.path.exists(lockfile) and pid:
        try:
            cf = init_cf(lockfile)
        except OSError:
            sys.stderr.write('lockfile points to an unknown connection file.\n')
            sys.stderr.write("Try 'kd5 stop'\n")
            sys.exit(1)
        if args.integer:
            message = 'Daemon is already running. Dropping argument {}\n'
            sys.stderr.write(message.format(args.integer))
            time.sleep(1.5)

    elif not os.path.exists(lockfile) and pid:
        no_lock_exit()

    elif args.integer == 'last' and not os.path.exists(lockfile):
        no_lock_exit()

    elif args.integer == 'last' and os.path.exists(lockfile):
        cmd = 'kd5 last'
        cf = with_daemon(lockfile, pidfile, cmd)

    elif args.integer:
        try:
            find_connection_file(str(args.integer))
        except OSError:
            message = '{}Error :{}\tCannot find kernel id. {} !\n\tExiting\n'
            sys.stderr.write(message.format(RED, RESET, args.integer))
            sys.exit(1)
        else:
            cmd = 'kd5 start ' + str(args.integer)
        cf = with_daemon(lockfile, pidfile, cmd)

    else:
        cmd = 'kd5 start'
        cf = with_daemon(lockfile, pidfile, cmd)

    return args, cf


def main(args=None):
    """ Launch cpyvke. """

    # Parse Config
    cfg = cfg_setup()
    config = cfg.run()

    # Define Paths
    logdir = os.path.expanduser('~') + '/.cpyvke/'
    lockfile = logdir + 'kd5.lock'
    pidfile = logdir + 'kd5.pid'
    logfile = logdir + 'cpyvke.log'

    # Logger
    logger = logging.getLogger("cpyvke")
    logger.setLevel(logging.DEBUG)

    # create the logging file handler
    handler = RotatingFileHandler(logfile, maxBytes=10*1024*1024,
                                  backupCount=5)
    logmsg = '%(asctime)s :: %(name)s :: %(threadName)s :: %(levelname)s :: %(message)s'
    formatter = logging.Formatter(logmsg, datefmt='%Y-%m-%d - %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # Parse arguments
    args, cf = parse_args(lockfile, pidfile)

    # Init kernel
    km, kc = connect_kernel(cf)

    # Init Curses App
    sock = SocketManager(config, logger)
    app = InitApp(kc, cf, config, sock)
    # Run App
    logger.info('cpyvke started')
    main_curse = MainWin(app, sock, logger)
    main_curse.display()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : jeu. 15 mars 2018 18:07:10 CET
# Last Modified : dim. 25 mars 2018 21:21:36 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import socket
from cpyvke.utils.comm import send_msg


class SocketManager:

    def __init__(self, config, logger):

        self.config = config
        self.logger = logger
        self.connected = False
        self.init_sockets()

    def init_main_socket(self):
        """ Init Main Socket. """

        try:
            hote = "localhost"
            sport = int(self.config['comm']['s-port'])
            self.MainSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.MainSock.connect((hote, sport))
            self.MainSock.setblocking(0)
            self.logger.debug('Connected to main socket')
        except ConnectionRefusedError:
            self.logger.error('Connection to stream socket failed ')

    def init_request_socket(self):
        """ Init Request Socket. """

        try:
            hote = "localhost"
            rport = int(self.config['comm']['r-port'])
            self.RequestSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.RequestSock.connect((hote, rport))
            self.RequestSock.setblocking(0)
            self.logger.debug('Connected to request socket')
        except Exception:
            self.logger.error('Connection to stream socket failed : \n', exc_info=True)

    def init_sockets(self):
        """ Init all sockets """

        self.init_main_socket()
        self.init_request_socket()

    def close_main_socket(self):
        """ Close Main socket. """

        try:
            self.MainSock.close()
            self.logger.debug('Main socket closed')
        except Exception:
            self.logger.error('Unable to close socket : ', exc_info=True)
            pass

    def close_request_socket(self):
        """ Close Request socket. """

        try:
            self.RequestSock.close()
            self.logger.debug('Request socket closed')
        except Exception:
            self.logger.error('Unable to close socket : ', exc_info=True)
            pass

    def close_sockets(self):
        """ Close all sockets """

        self.close_main_socket()
        self.close_request_socket()

    def restart_sockets(self):
        """ Stop then start connection to sockets. """

        self.close_sockets()
        self.init_sockets()

    def check_main_socket(self):
        """ Test if connection to daemon is alive. """

        try:
            send_msg(self.MainSock, '<TEST>')
            self.connected = True
        except BlockingIOError:
            self.connected = True
        except OSError:
            self.connected = False

    def warning_socket(self, wng):
        """ Check connection and display warning. """

        self.check_main_socket()
        if self.connected:
            wng.display('  Connected to socket  ')
        else:
            wng.display(' Disconnected from socket ')

    def force_update(self, wng):
        """ Force update of variable list sending fake code to Daemon """

        send_msg(self.RequestSock, '<code> ')
        wng.display('Reloading Variable List...')

    def del_var(self, varname, wng):
        """ Delete a variable from kernel. """

        code = 'del {}'.format(varname)
        try:
            send_msg(self.RequestSock, '<code>' + code)
            self.logger.debug("Send delete signal for variable {}".format(varname))
        except Exception:
            self.logger.error('Delete variable :', exc_info=True)
            wng.display('Kernel Busy ! Try again...')
        else:
            wng.display('Variable {} deleted'.format(varname))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : mar. 29 nov. 2016 23:18:27 CET
# Last Modified : lun. 09 avril 2018 22:46:28 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


import sys
import os
from time import sleep
try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser


class cfg_setup:
    """ Handle configuration file. """

    def __init__(self):

        self.cfg = ConfigParser.RawConfigParser()
        self.home = os.path.expanduser("~")
        self.path = self.home + '/.cpyvke/'
        # Check if config dir exists. If not create it.
        self.check_dir(self.path)

    @staticmethod
    def check_dir(directory):
        """ Check if dir exists. If not, create it."""

        if not os.path.isdir(directory):
            os.makedirs(directory)
            print("Create directory :", directory)
            sleep(0.5)

    def init_cfg(self):
        """ Check if cpyvke.conf exists. If not create it. """

        if os.path.exists(self.path + 'cpyvke.conf') is False:
            open(self.path + 'cpyvke.conf', 'a').close()
            print("Create configuration file : {}cpyvke.conf".format(self.path))
            sleep(0.5)
            self.write_config()

    def write_config(self):
        """ Write configuration file. """

        self.cfg.add_section('main colors')
        self.cfg.set('main colors', 'title', 'white, transparent')
        self.cfg.set('main colors', 'text', 'white, transparent')
        self.cfg.set('main colors', 'highlight', 'black, cyan')
        self.cfg.set('main colors', 'border', 'white, transparent')
        self.cfg.set('main colors', 'ascii', 'red, transparent')

        self.cfg.add_section('explorer colors')
        self.cfg.set('explorer colors', 'title', 'cyan, transparent')
        self.cfg.set('explorer colors', 'text', 'white, transparent')
        self.cfg.set('explorer colors', 'highlight', 'black, cyan')
        self.cfg.set('explorer colors', 'border', 'cyan, transparent')

        self.cfg.add_section('kernel colors')
        self.cfg.set('kernel colors', 'title', 'red, transparent')
        self.cfg.set('kernel colors', 'text', 'white, transparent')
        self.cfg.set('kernel colors', 'highlight', 'white, transparent')
        self.cfg.set('kernel colors', 'border', 'red, transparent')
        self.cfg.set('kernel colors', 'connected', 'green, transparent')
        self.cfg.set('kernel colors', 'died', 'red, transparent')
        self.cfg.set('kernel colors', 'alive', 'cyan, transparent')

        self.cfg.add_section('bar colors')
        self.cfg.set('bar colors', 'kernel', 'white, transparent')
        self.cfg.set('bar colors', 'connected', 'green, transparent')
        self.cfg.set('bar colors', 'disconnected', 'red, transparent')
        self.cfg.set('bar colors', 'help', 'white, transparent')

        self.cfg.add_section('warning colors')
        self.cfg.set('warning colors', 'text', 'red, transparent')
        self.cfg.set('warning colors', 'border', 'red, transparent')

        self.cfg.add_section('path')
        self.cfg.set('path', 'save-dir', '$HOME/.cpyvke/save/')

        self.cfg.add_section('font')
        self.cfg.set('font', 'powerline-font', 'False')
        self.cfg.set('font', 'ascii-font', 'small')

        self.cfg.add_section('comm')
        self.cfg.set('comm', 's-port', 15555)
        self.cfg.set('comm', 'r-port', 15556)

        self.cfg.add_section('daemon')
        self.cfg.set('daemon', 'refresh', 0.1)

        self.cfg.add_section('kernel version')
        self.cfg.set('kernel version', 'version', '3')

        with open(self.path + 'cpyvke.conf', 'w') as configfile:
            self.cfg.write(configfile)

    def run(self):
        """ Run configuration. """

        try:
            # Check if config file exist. If not create it
            self.init_cfg()
            # Read config file
            self.cfg.read(self.path + 'cpyvke.conf')

            # save_dir
            if self.cfg.has_option('path', 'save-dir'):
                self.save_dir = self.cfg.get('path', 'save-dir')
                self.save_dir = self.save_dir.replace('$HOME', self.home)
                if self.save_dir[-1] != '/':
                    self.save_dir = self.save_dir + '/'
            else:
                self.save_dir = self.path + 'save/'

            # FONT
            if self.cfg.has_option('font', 'powerline-font'):
                pwf = self.cfg.get('font', 'powerline-font')
            else:
                pwf = 'False'

            if self.cfg.has_option('font', 'ascii-font'):
                ascii = self.cfg.get('font', 'ascii-font')
            else:
                ascii = 'small'

            # KERNEL VERSION
            if self.cfg.has_option('kernel version', 'version'):
                kver = self.cfg.get('kernel version', 'version')
            else:
                kver = 3

            # Refresh delay
            if self.cfg.has_option('daemon', 'refresh'):
                delay = self.cfg.get('daemon', 'refresh')
            else:
                delay = 0.1

            # COMM
            if self.cfg.has_option('comm', 'r-port'):
                rport = self.cfg.get('comm', 'r-port')
            else:
                rport = 15556

            if self.cfg.has_option('comm', 's-port'):
                sport = self.cfg.get('comm', 's-port')
            else:
                sport = 15555

            # WARNING COLORS
            if self.cfg.has_option('warning colors', 'text'):
                wg_txt = self.cfg.get('warning colors', 'text')
            else:
                wg_txt = "red, transparent"

            if self.cfg.has_option('warning colors', 'border'):
                wg_bdr = self.cfg.get('warning colors', 'border')
            else:
                wg_bdr = "red, transparent"

            # BAR COLORS
            if self.cfg.has_option('bar colors', 'kernel'):
                br_kn = self.cfg.get('bar colors', 'kernel')
            else:
                br_kn = "white, transparent"

            if self.cfg.has_option('bar colors', 'help'):
                br_hlp = self.cfg.get('bar colors', 'help')
            else:
                br_hlp = "white, transparent"

            if self.cfg.has_option('bar colors', 'connected'):
                br_co = self.cfg.get('bar colors', 'connected')
            else:
                br_co = "green, transparent"

            if self.cfg.has_option('bar colors', 'disconnected'):
                br_dco = self.cfg.get('bar colors', 'disconnected')
            else:
                br_dco = "red, transparent"

            # MAIN COLORS
            if self.cfg.has_option('main colors', 'text'):
                main_txt = self.cfg.get('main colors', 'text')
            else:
                main_txt = "white, transparent"

            if self.cfg.has_option('main colors', 'border'):
                main_bdr = self.cfg.get('main colors', 'border')
            else:
                main_bdr = "white, transparent"

            if self.cfg.has_option('main colors', 'title'):
                main_ttl = self.cfg.get('main colors', 'title')
            else:
                main_ttl = "white, transparent"

            if self.cfg.has_option('main colors', 'highlight'):
                main_hh = self.cfg.get('main colors', 'highlight')
            else:
                main_hh = "black, cyan"

            if self.cfg.has_option('main colors', 'ascii'):
                main_asc = self.cfg.get('main colors', 'ascii')
            else:
                main_asc = "red, transparent"

            # EXPLORER COLORS
            if self.cfg.has_option('explorer colors', 'text'):
                exp_txt = self.cfg.get('explorer colors', 'text')
            else:
                exp_txt = "white, transparent"

            if self.cfg.has_option('explorer colors', 'border'):
                exp_bdr = self.cfg.get('explorer colors', 'border')
            else:
                exp_bdr = "white, transparent"

            if self.cfg.has_option('explorer colors', 'title'):
                exp_ttl = self.cfg.get('explorer colors', 'title')
            else:
                exp_ttl = "white, transparent"

            if self.cfg.has_option('explorer colors', 'highlight'):
                exp_hh = self.cfg.get('explorer colors', 'highlight')
            else:
                exp_hh = "black, cyan"

            # KERNEL COLORS
            if self.cfg.has_option('kernel colors', 'text'):
                kn_txt = self.cfg.get('kernel colors', 'text')
            else:
                kn_txt = "white, transparent"

            if self.cfg.has_option('kernel colors', 'border'):
                kn_bdr = self.cfg.get('kernel colors', 'border')
            else:
                kn_bdr = "white, transparent"

            if self.cfg.has_option('kernel colors', 'title'):
                kn_ttl = self.cfg.get('kernel colors', 'title')
            else:
                kn_ttl = "white, transparent"

            if self.cfg.has_option('kernel colors', 'highlight'):
                kn_hh = self.cfg.get('kernel colors', 'highlight')
            else:
                kn_hh = "black, red"

            if self.cfg.has_option('kernel colors', 'connected'):
                kn_co = self.cfg.get('kernel colors', 'connected')
            else:
                kn_co = "green, transparent"

            if self.cfg.has_option('kernel colors', 'alive'):
                kn_al = self.cfg.get('kernel colors', 'alive')
            else:
                kn_al = "cyan, transparent"

            if self.cfg.has_option('kernel colors', 'died'):
                kn_di = self.cfg.get('kernel colors', 'died')
            else:
                kn_di = "red, transparent"

            # Colors :
            self.Config = {'mn': {'txt': main_txt,
                                  'bdr': main_bdr,
                                  'ttl': main_ttl,
                                  'asc': main_asc,
                                  'hh': main_hh},
                           'xp': {'txt': exp_txt,
                                  'bdr': exp_bdr,
                                  'ttl': exp_ttl,
                                  'hh': exp_hh},
                           'kn': {'txt': kn_txt,
                                  'bdr': kn_bdr,
                                  'ttl': kn_ttl,
                                  'hh': kn_hh,
                                  'co': kn_co,
                                  'di': kn_di,
                                  'al': kn_al},
                           'wg': {'txt': wg_txt,
                                  'bdr': wg_bdr},
                           'br': {'kn': br_kn,
                                  'hlp': br_hlp,
                                  'co': br_co,
                                  'dco': br_dco},
                           'path': {'save-dir': self.save_dir},
                           'font': {'pw-font': pwf,
                                    'ascii-font': ascii},
                           'kernel version': {'version': kver},
                           'comm': {'s-port': sport,
                                    'r-port': rport},
                           'daemon': {'refresh': delay}}

            # Init save Directory
            self.check_dir(self.save_dir)

            return self.Config

        except ConfigParser.Error as err:
            print('Bad cfg file : ', err)
            sys.exit(1)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Mon Nov 14 09:08:25 2016
# Last Modified : mar. 03 avril 2018 11:12:09 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import curses
from curses import panel


def cmd_lst():
    """ Prompt command list """

    lst = ['kernel-manager',
           'daemon-restart-connection', 'daemon-connect', 'daemon-disconnect',
           'daemon-restart',
           'help',
           'variable-explorer',
           'toggle-debug']
    lst.sort()

    return lst


class Prompt:
    """ Prompt class.

    Prompt.simple : simple prompt that returns a user input
    Prompt.with_completion : Prompt with completion capabitilies that returns a user input
    Prompt.panel : a simple prompt in a panel that returns a user input

    """

    def __init__(self, app):
        """ Class constructor """

        # Arguments
        self.app = app
        self.config = app.config

        # Init input
        self.usr_input = ''

        # Screen
        self.stdscr = app.stdscr

    def simple(self, prefix):
        """ Display prompt and return user input. """

        # Reinit input
        self.usr_input = ''

        # Enable echoing of characters
        curses.echo()
        self.app.stdscr.attrset(self.app.c_main_txt | curses.A_BOLD)
        # Erase before ! (overwrite with spaces)
        self.app.stdscr.addstr(self.app.screen_height-1, 0,
                               ' '*(self.app.screen_width-1),
                               curses.A_DIM | self.app.c_main_txt)

        self.app.stdscr.addstr(self.app.screen_height-1, 0, prefix, curses.A_DIM | self.app.c_main_txt)
        self.usr_input = self.app.stdscr.getstr(self.app.screen_height-1, len(prefix),
                                                self.app.screen_width - len(prefix) - 1).decode('utf-8')
        # Disable echoing of characters
        curses.noecho()
        # Restore color of the main window
        self.app.stdscr.attrset(self.app.c_main_bdr | curses.A_BOLD)

        return self.usr_input

    def with_completion(self, prefix):
        """ Display prompt with completion and return user input. """

        # Reinit input
        self.usr_input = ''
        self.app.stdscr.attrset(self.app.c_main_txt | curses.A_BOLD)
        # Enable echoing of characters
        curses.echo()
        curses.halfdelay(1)

        mkey = ''
        pkey = -1
        while pkey not in ['\n']:    # ESC/ENTER

            try:
                # get_wch return a character for most keys, or an integer
                # for function keys, keypad keys, and other special keys
                if not mkey:
                    pkey = self.app.stdscr.get_wch()

                else:
                    pkey = mkey
                    mkey = ''

                if pkey in ['^[', 27, curses.KEY_RESIZE]:
                    self.usr_input = ''
                    break

                elif pkey in ['\t']:    # TAB
                    match = [cmd for cmd in cmd_lst() if cmd.startswith(self.usr_input)]

                    if not match:
                        pass
                    elif len(match) == 1:
                        self.usr_input = match[0]
                    else:
                        mkey = self._complete(match, prefix)

                elif pkey in [263]:    # DEL
                    self.usr_input = self.usr_input[:-1]
                    if not self.usr_input:
                        break

                elif type(pkey) is not int and len(self.usr_input + pkey.rstrip()) < self.app.screen_width - 2:
                    self.usr_input = self.usr_input + pkey.rstrip()

            except curses.error:
                pass

            # Erase before ! (overwrite with space characters)
            self.app.stdscr.addstr(self.app.screen_height-1, 0,
                                   ' '*(self.app.screen_width-1),
                                   curses.A_BOLD | self.app.c_main_txt)

            self.app.stdscr.addstr(self.app.screen_height-1, 0,
                                   prefix + self.usr_input,
                                   curses.A_BOLD | self.app.c_main_txt)

        # Restore color of the main window and switch of echo !
        curses.noecho()
        self.app.stdscr.attrset(self.app.c_main_bdr | curses.A_BOLD)

        return self.usr_input

    def _complete(self, match, prefix):
        """ Completion method """

        mkey = -1
        i = 0
        proposal = match[0]
        while mkey not in ['\n']:
            try:
                mkey = self.app.stdscr.get_wch()
                if mkey in ['\t'] and i < len(match) - 1:
                    i += 1
                    proposal = match[i]
                    if i == len(match) - 1:
                        i = -1
                elif mkey not in ['\t', '\n']:
                    return mkey
            except curses.error:
                pass

            self.app.stdscr.addstr(self.app.screen_height-1, 0,
                                   ' '*(self.app.screen_width-1),
                                   curses.A_BOLD | self.app.c_main_txt)
            self.app.stdscr.addstr(self.app.screen_height-1, 0,
                                   prefix + proposal, curses.A_DIM | self.app.c_main_txt)
            self.app.stdscr.addstr(self.app.screen_height-1, 0,
                                   prefix + self.usr_input, curses.A_BOLD | self.app.c_exp_txt)

        self.usr_input = match[i]
        return mkey

    def display(self, msg):
        """ display message in prompt. """

        self.app.stdscr.addstr(self.app.screen_height-1, 0,
                               msg, curses.A_BOLD | self.app.c_warn_txt)

    def panel(self, txt_msg, win_title):
        """ Prompt on a dedicated panel """

        # Init Menu
        self.win_title = win_title
        iwin = self.app.stdscr.subwin(self.app.panel_height, self.app.screen_width, 0, 0)
        iwin.keypad(1)

        # Send menu to a panel
        ipan = panel.new_panel(iwin)

        ipan.top()        # Push the panel to the bottom of the stack.
        ipan.show()       # Display the panel (which might have been hidden)
        iwin.clear()
        iwin.bkgd(self.app.c_main_txt)
        iwin.attrset(self.app.c_main_bdr | curses.A_BOLD)  # Change border color
        iwin.border(0)
        if self.app.config['font']['pw-font'] == 'True':
            iwin.addstr(0, int((self.app.screen_width-len(self.win_title))/2),
                        '', self.app.c_main_pwf)
            iwin.addstr(self.win_title, self.app.c_main_ttl | curses.A_BOLD)
            iwin.addstr('', self.app.c_main_pwf)
        else:
            iwin.addstr(0, int((self.app.screen_width-len(self.win_title))/2),
                        '| ' + self.win_title + ' |', self.app.c_main_ttl | curses.A_BOLD)

        curses.echo()
        iwin.addstr(2, 3, txt_msg, curses.A_BOLD | self.app.c_main_txt)
        self.usr_input = iwin.getstr(2, len(txt_msg) + 4,
                                     self.app.screen_width - len(txt_msg) - 8).decode('utf-8')
        curses.noecho()
        ipan.hide()

        return self.usr_input



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : sam. 17 mars 2018 11:22:56 CET
# Last Modified : lun. 09 avril 2018 22:37:36 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""


def ascii_cpyvke(font='small'):
    """ figlet -f font "Cpyvke" """

    if font == 'small':
        return ["  ___               _       ",
                " / __|_ __ _  ___ _| |_____ ",
                "| (__| '_ \ || \ V / / / -_)",
                " \___| .__/\_, |\_/|_\_\___|  v1.2.9",
                "     |_|   |__/             "]
    elif font == 'smshadow':
        return ["   __|              |        ",
                "  (    _ \ |  |\ \ /| /  -_) ",
                " \___|.__/\_, | \_/_\_\\___|  v1.2.9",
                "     _|   ___/               "]
    elif font == 'pagga':
        return ["░█▀▀░█▀█░█░█░█░█░█░█░█▀▀",
                "░█░░░█▀▀░░█░░▀▄▀░█▀▄░█▀▀",
                "░▀▀▀░▀░░░░▀░░░▀░░▀░▀░▀▀▀  v1.2.9"]
    else:
        return ["Ascii font not found !"]




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2018 Cyril Desjouy <ipselium@free.fr>
#
# This file is part of cpyvke
#
# cpyvke is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cpyvke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cpyvke. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : Wed Nov 9 10:03:04 2016
# Last Modified : lun. 09 avril 2018 22:51:37 CEST
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import time
import psutil
import locale
import traceback
import curses

from cpyvke.utils.colors import Colors
from cpyvke.utils.comm import send_msg
from cpyvke.utils.display import str_reduce

locale.setlocale(locale.LC_ALL, '')


def check_size(function):
    """ Decorator to check size of the curses window.
    Blank screen and display a warning if size of the terminal is too small."""

    def wrapper(*args, **kwargs):

        # Fetch self instance which is 1st element of args
        app = args[0].app

        # Min terminal size allowed
        term_min_height = 15
        term_min_width = 70

        # Blank screen if dimension are too small
        if app.screen_height < term_min_height or app.screen_width < term_min_width:
            # Messages
            msg_actual = '{}x{}'.format(app.screen_width, app.screen_height)
            msg_limit = 'Win must be > {}x{}'.format(term_min_width, term_min_height)
            # Message locations
            y_mid = int(app.screen_height/2)
            x_mid_limit = int((app.screen_width-len(msg_limit))/2)
            x_mid_actual = int((app.screen_width-len(msg_actual))/2)
            # Update curses
            app.stdscr.erase()
            try:
                app.stdscr.addstr(y_mid, x_mid_limit, msg_limit, app.c_warn_txt | curses.A_BOLD)
                app.stdscr.addstr(y_mid+1, x_mid_actual, msg_actual, app.c_warn_txt | curses.A_BOLD)
            except curses.error:
                pass
            app.stdscr.refresh()
            time.sleep(0.2)
        else:
            function(*args, **kwargs)

    return wrapper


class InitApp:
    """ Initlication. """

    def __init__(self, kc, cf, config, sock):
        """  """

        # Arguments
        self.kc = kc
        self.cf = cf
        self.config = config
        self.sock = sock

        # Init CUI :
        self.close_signal = 'continue'
        self.stdscr = curses.initscr()   # Init curses

        # Ressources
        self.ressources = psutil.Process(os.getpid())

        # Curses options
        self.stdscr.keypad(1)            #
        self.curse_delay = 10            # 1s timer on each getch
        curses.noecho()             # Wont print the input og getch (keys)
        curses.cbreak()             #
        curses.curs_set(0)          #
        # How many tenths of second are waited to refresh screen, from 1 to 255
        curses.halfdelay(self.curse_delay)
        # Init color pairs
        Colors(self.config)
        # Assign color pairs to variables
        self.color_def()
        # Set some colors
        self.stdscr.bkgd(self.c_main_txt)

        # Some variables
        self.debug = False
        self.kernel_change = False
        self.explorer_switch = False
        self.kernel_switch = False
        self.var_nb = 0

    @property
    def screen_width(self):
        _, screen_width = self.stdscr.getmaxyx()
        return screen_width

    @property
    def screen_height(self):
        screen_height, _ = self.stdscr.getmaxyx()
        return screen_height

    @property
    def panel_height(self):
        return self.screen_height - 2

    @property
    def row_max(self):
        return self.panel_height - 2

    @property
    def kdown(self):
        """ Down keys """
        return [curses.KEY_DOWN, 258, 106]

    @property
    def kup(self):
        """ Up keys """
        return [curses.KEY_UP, 259, 107]

    @property
    def kleft(self):
        """ Left keys """
        return [curses.KEY_LEFT, 260, 104, 339]

    @property
    def kright(self):
        """ Right keys """
        return [curses.KEY_RIGHT, 261, 108, 338]

    @property
    def kenter(self):
        """ Enter keys """
        return [curses.KEY_ENTER, ord("\n"), 10, 32]

    @property
    def kquit(self):
        """ quit keys """
        return [27, 113]

    def color_def(self):
        """ Definition of all color variables """

        self.c_warn_txt = curses.color_pair(1)
        self.c_warn_bdr = curses.color_pair(2)

        self.c_main_txt = curses.color_pair(11)
        self.c_main_bdr = curses.color_pair(12)
        self.c_main_ttl = curses.color_pair(13)
        self.c_main_hh = curses.color_pair(14)
        self.c_main_pwf = curses.color_pair(15)
        self.c_main_asc = curses.color_pair(16)

        self.c_exp_txt = curses.color_pair(21)
        self.c_exp_bdr = curses.color_pair(22)
        self.c_exp_ttl = curses.color_pair(23)
        self.c_exp_hh = curses.color_pair(24)
        self.c_exp_pwf = curses.color_pair(25)

        self.c_kern_txt = curses.color_pair(31)
        self.c_kern_bdr = curses.color_pair(32)
        self.c_kern_ttl = curses.color_pair(33)
        self.c_kern_hh = curses.color_pair(34)
        self.c_kern_pwf = curses.color_pair(38)

        self.c_kern_co = curses.color_pair(35)
        self.c_kern_al = curses.color_pair(36)
        self.c_kern_di = curses.color_pair(37)

        self.c_bar_kn = curses.color_pair(41)
        self.c_bar_hlp = curses.color_pair(42)
        self.c_bar_co = curses.color_pair(43)
        self.c_bar_dco = curses.color_pair(44)
        self.c_bar_kn_pwf = curses.color_pair(45)
        self.c_bar_hlp_pwf = curses.color_pair(46)
        self.c_bar_kn_pwfi = curses.color_pair(47)
        self.c_bar_kn_pwfk = curses.color_pair(48)
        self.c_bar_kn_pwfc = curses.color_pair(49)
        self.c_bar_kn_pwfd = curses.color_pair(40)

    def dbg_pad(self, pkey, search=None, filt=None, mk_sort=None):
        """ Display debug informations """

        pad_width = 19

        lst = [' cpu: {}%'.format(self.ressources.cpu_percent()),
               ' mem: {:.2f}%'.format(self.ressources.memory_percent()),
               ' size : {}x{}'.format(self.screen_width, self.screen_height),
               ' colors : {}'.format(curses.COLORS),
               ' key : {}'.format(pkey),
               ' sort : {}'.format(mk_sort),
               str_reduce(' search : {}'.format(search), pad_width - 3),
               str_reduce(' limit : {}'.format(filt), pad_width - 3)]

        # Init Menu
        pad_height = len(lst) + 2
        pad = self.stdscr.subwin(pad_height, pad_width, 2, int((self.screen_width-pad_width+1)/2))
        pad.attrset(self.c_warn_bdr | curses.A_BOLD)    # change border color
        pad.keypad(1)
        pad.clear()
        pad.border(0)

        # Fill pad
        for idx, item in enumerate(lst):
            if 'None' not in item:
                pad.addstr(idx + 1, 2, item, curses.A_DIM | self.c_main_txt)

        pad.refresh()

    def status_bar(self):
        """ Check and display kernel informations """

        debug_info_id = 'kernel {} '.format(self.cf.split('-')[1].split('.')[0])
        debug_info_obj = '{} obj.'.format(self.var_nb)

        # Kernel Info
        if self.config['font']['pw-font'] == 'True':
            self.stdscr.addstr(self.screen_height-2, 0, ' ', self.c_bar_co | curses.A_BOLD)
            self.stdscr.addstr('', self.c_bar_kn_pwfi | curses.A_BOLD)
            self.stdscr.addstr(' Daemon ', self.c_bar_kn)
            self.stdscr.addstr('', self.c_bar_kn_pwfk | curses.A_BOLD)
            if self.sock.connected:
                self.stdscr.addstr(' connected ', self.c_bar_co | curses.A_BOLD)
                self.stdscr.addstr(' ', self.c_bar_kn_pwfc | curses.A_BOLD)
                self.stdscr.addstr(debug_info_id, self.c_bar_kn)
                self.stdscr.addstr(' ', self.c_bar_kn | curses.A_BOLD)
                self.stdscr.addstr(debug_info_obj, self.c_bar_kn)
                self.stdscr.addstr('', self.c_bar_kn_pwf | curses.A_BOLD)
            else:
                self.stdscr.addstr(' disconnected ', self.c_bar_dco | curses.A_BOLD)
                self.stdscr.addstr('', self.c_bar_kn_pwfd | curses.A_BOLD)

        else:
            self.stdscr.addstr(self.screen_height-2, 0, '< Kernel : ',
                               self.c_bar_kn | curses.A_BOLD)
            if self.kc.is_alive():
                self.stdscr.addstr('connected', self.c_bar_co | curses.A_BOLD)
                self.stdscr.addstr(' [' + debug_info_id + debug_info_obj + ']',
                                   self.c_bar_kn | curses.A_BOLD)
            else:
                self.stdscr.addstr('disconnected ', self.c_bar_dco | curses.A_BOLD)
            self.stdscr.addstr(' >', self.c_bar_kn | curses.A_BOLD)

        # Help
        if self.config['font']['pw-font'] == 'True':
            self.stdscr.addstr(self.screen_height-2, self.screen_width-9,
                               '', self.c_bar_hlp_pwf | curses.A_BOLD)
            self.stdscr.addstr(' ?:help ', self.c_bar_hlp | curses.A_BOLD)
        else:
            self.stdscr.addstr(self.screen_height-2, self.screen_width-10,
                               '< ?:help >', self.c_bar_hlp | curses.A_BOLD)

    def shutdown(self):
        """ Shutdown CUI, Daemon, and kernel """

        curses.endwin()

        if self.close_signal == 'close':
            print('Exiting ! Closing cpyvke...')
        elif self.close_signal == 'shutdown':
            print('Exiting ! Shutting down daemon...')
            send_msg(self.sock.RequestSock, '<_stop>')

        self.kc.stop_channels()
        self.sock.close_sockets()
        self.kill_all_figs()   # Stop all figure subprocesses

    @staticmethod
    def kill_all_figs():
        """ Kill all figures (running in different processes) """

        import multiprocessing

        if len(multiprocessing.active_children()) == 1:
            print('{} figure killed'.format(len(multiprocessing.active_children())))
        elif len(multiprocessing.active_children()) > 1:
            print('{} figures killed'.format(len(multiprocessing.active_children())))

        for child in multiprocessing.active_children():
            child.terminate()

    def exit_with_error(self):
        """ If error, send terminate signal to daemon and resore terminal to
            sane state """

        self.close_signal = 'close'
        self.stdscr.keypad(0)
        curses.echo()
        curses.nocbreak()
        curses.endwin()
        self.shutdown()
        traceback.print_exc()           # Print the exception
